@@ -26,6 +26,7 @@
 #include <stdio.h>
 
 #include "libs/fvwmlib.h"
+#include "libs/FScreen.h"
 #include "fvwm.h"
 #include "externs.h"
 #include "cursor.h"
@@ -35,6 +36,7 @@
 #include "screen.h"
 #include "update.h"
 #include "style.h"
+#include "focus.h"
 
 
 /**********************************************************************
@@ -67,14 +69,17 @@ char *CreateFlagString(char *string, char **restptr)
     start = c;
     length = 0;
     while (*c != closeopt) {
+      char *d;
       if (*c == 0) {
 	fvwm_msg(ERR, "CreateFlagString",
 		 "Conditionals require closing parenthesis");
 	*restptr = NULL;
 	return NULL;
       }
-      c++;
-      length++;
+      /* skip quoted string */
+      d = SkipQuote(c, NULL, NULL, NULL);
+      length += d - c;
+      c = d;
     }
 
     /* We must allocate a new string because we null terminate the string
@@ -137,100 +142,129 @@ void CreateConditionMask(char *flags, WindowConditionMask *mask)
 
   while (condition)
   {
-    if (StrEquals(condition,"Iconic"))
+    if (StrEquals(condition,"AcceptsFocus"))
+    {
+      mask->my_flags.do_accept_focus = 1;
+      mask->my_flags.use_do_accept_focus = 1;
+    }
+    else if (StrEquals(condition,"!AcceptsFocus"))
+    {
+      mask->my_flags.do_accept_focus = 0;
+      mask->my_flags.use_do_accept_focus = 1;
+    }
+    else if (StrEquals(condition,"Iconic"))
     {
       SET_ICONIFIED(mask, 1);
       SETM_ICONIFIED(mask, 1);
     }
-    else if(StrEquals(condition,"!Iconic"))
+    else if (StrEquals(condition,"!Iconic"))
     {
       SET_ICONIFIED(mask, 0);
       SETM_ICONIFIED(mask, 1);
     }
-    else if(StrEquals(condition,"Visible"))
+    else if (StrEquals(condition,"Visible"))
     {
       SET_PARTIALLY_VISIBLE(mask, 1);
       SETM_PARTIALLY_VISIBLE(mask, 1);
     }
-    else if(StrEquals(condition,"!Visible"))
+    else if (StrEquals(condition,"!Visible"))
     {
       SET_PARTIALLY_VISIBLE(mask, 0);
       SETM_PARTIALLY_VISIBLE(mask, 1);
     }
-    else if(StrEquals(condition,"PlacedByButton3"))
+    else if (StrEquals(condition,"PlacedByButton3"))
     {
       SET_PLACED_WB3(mask, 1);
       SETM_PLACED_WB3(mask, 1);
     }
-    else if(StrEquals(condition,"!PlacedByButton3"))
+    else if (StrEquals(condition,"!PlacedByButton3"))
     {
       SET_PLACED_WB3(mask, 0);
       SETM_PLACED_WB3(mask, 1);
     }
-    else if(StrEquals(condition,"Raised"))
+    else if (StrEquals(condition,"Raised"))
     {
       SET_FULLY_VISIBLE(mask, 1);
       SETM_FULLY_VISIBLE(mask, 1);
     }
-    else if(StrEquals(condition,"!Raised"))
+    else if (StrEquals(condition,"!Raised"))
     {
       SET_FULLY_VISIBLE(mask, 0);
       SETM_FULLY_VISIBLE(mask, 1);
     }
-    else if(StrEquals(condition,"Sticky"))
+    else if (StrEquals(condition,"Sticky"))
     {
       SET_STICKY(mask, 1);
       SETM_STICKY(mask, 1);
     }
-    else if(StrEquals(condition,"!Sticky"))
+    else if (StrEquals(condition,"!Sticky"))
     {
       SET_STICKY(mask, 0);
       SETM_STICKY(mask, 1);
     }
-    else if(StrEquals(condition,"Maximized"))
+    else if (StrEquals(condition,"Maximized"))
     {
       SET_MAXIMIZED(mask, 1);
       SETM_MAXIMIZED(mask, 1);
     }
-    else if(StrEquals(condition,"!Maximized"))
+    else if (StrEquals(condition,"!Maximized"))
     {
       SET_MAXIMIZED(mask, 0);
       SETM_MAXIMIZED(mask, 1);
     }
-    else if(StrEquals(condition,"Shaded"))
+    else if (StrEquals(condition,"Shaded"))
     {
       SET_SHADED(mask, 1);
       SETM_SHADED(mask, 1);
     }
-    else if(StrEquals(condition,"!Shaded"))
+    else if (StrEquals(condition,"!Shaded"))
     {
       SET_SHADED(mask, 0);
       SETM_SHADED(mask, 1);
     }
-    else if(StrEquals(condition,"Transient"))
+    else if (StrEquals(condition,"Transient"))
     {
       SET_TRANSIENT(mask, 1);
       SETM_TRANSIENT(mask, 1);
     }
-    else if(StrEquals(condition,"!Transient"))
+    else if (StrEquals(condition,"!Transient"))
     {
       SET_TRANSIENT(mask, 0);
       SETM_TRANSIENT(mask, 1);
     }
-    else if(StrEquals(condition,"CurrentDesk"))
+    else if (StrEquals(condition,"PlacedByFvwm"))
+    {
+      SET_PLACED_BY_FVWM(mask, 1);
+      SETM_PLACED_BY_FVWM(mask, 1);
+    }
+    else if (StrEquals(condition,"!PlacedByFvwm"))
+    {
+      SET_PLACED_BY_FVWM(mask, 0);
+      SETM_PLACED_BY_FVWM(mask, 1);
+    }
+    else if (StrEquals(condition,"CurrentDesk"))
       mask->my_flags.needs_current_desk = 1;
-    else if(StrEquals(condition,"CurrentPage"))
+    else if (StrEquals(condition,"CurrentPage"))
     {
       mask->my_flags.needs_current_desk = 1;
       mask->my_flags.needs_current_page = 1;
     }
-    else if(StrEquals(condition,"CurrentPageAnyDesk") ||
+    else if (StrEquals(condition,"CurrentGlobalPage"))
+    {
+      mask->my_flags.needs_current_desk = 1;
+      mask->my_flags.needs_current_global_page = 1;
+    }
+    else if (StrEquals(condition,"CurrentPageAnyDesk") ||
 	    StrEquals(condition,"CurrentScreen"))
       mask->my_flags.needs_current_page = 1;
-    else if(StrEquals(condition,"CirculateHit"))
+    else if (StrEquals(condition,"CurrentGlobbalPageAnyDesk"))
+      mask->my_flags.needs_current_global_page = 1;
+    else if (StrEquals(condition,"CirculateHit"))
       mask->my_flags.use_circulate_hit = 1;
-    else if(StrEquals(condition,"CirculateHitIcon"))
+    else if (StrEquals(condition,"CirculateHitIcon"))
       mask->my_flags.use_circulate_hit_icon = 1;
+    else if (StrEquals(condition,"CirculateHitShaded"))
+      mask->my_flags.use_circulate_hit_shaded = 1;
     else if (StrEquals(condition, "Layer"))
     {
        if (sscanf(tmp,"%d",&mask->layer))
@@ -244,7 +278,7 @@ void CreateConditionMask(char *flags, WindowConditionMask *mask)
           mask->layer = -1; /* needs current layer */
        }
     }
-    else if(!mask->my_flags.needs_name && !mask->my_flags.needs_not_name)
+    else if (!mask->my_flags.needs_name && !mask->my_flags.needs_not_name)
     {
       /* only 1st name to avoid mem leak */
       mask->name = condition;
@@ -262,10 +296,10 @@ void CreateConditionMask(char *flags, WindowConditionMask *mask)
       free(prev_condition);
 
     prev_condition = condition;
-    tmp = GetNextToken(tmp, &condition);
+    tmp = GetNextSimpleOption(tmp, &condition);
   }
 
-  if(prev_condition)
+  if (prev_condition)
     free(prev_condition);
 }
 
@@ -281,6 +315,7 @@ Bool MatchesConditionMask(FvwmWindow *fw, WindowConditionMask *mask)
   Bool fMatchesClass;
   Bool fMatchesResource;
   Bool fMatches;
+  FvwmWindow *sf = get_focus_window();
 
   if (!blockcmpmask((char *)&(fw->flags), (char *)&(mask->flags),
                     (char *)&(mask->flag_mask), sizeof(fw->flags)))
@@ -297,13 +332,11 @@ Bool MatchesConditionMask(FvwmWindow *fw, WindowConditionMask *mask)
   if (!mask->my_flags.use_circulate_hit && DO_SKIP_CIRCULATE(fw))
     return 0;
 
-  /* This logic looks terribly wrong to me, but it was this way before so I
-   * did not change it (domivogt (24-Dec-1998)) */
-  if (!DO_SKIP_ICON_CIRCULATE(mask) && IS_ICONIFIED(fw) &&
+  if (!mask->my_flags.use_circulate_hit_icon && IS_ICONIFIED(fw) &&
       DO_SKIP_ICON_CIRCULATE(fw))
     return 0;
 
-  if (!DO_SKIP_SHADED_CIRCULATE(mask) && IS_SHADED(fw) &&
+  if (!mask->my_flags.use_circulate_hit_shaded && IS_SHADED(fw) &&
       DO_SKIP_SHADED_CIRCULATE(fw))
     return 0;
 
@@ -312,10 +345,24 @@ Bool MatchesConditionMask(FvwmWindow *fw, WindowConditionMask *mask)
 
   if (mask->my_flags.needs_current_desk && fw->Desk != Scr.CurrentDesk)
     return 0;
-
-  if (mask->my_flags.needs_current_page &&
-      !IsRectangleOnThisPage(&(fw->frame_g), Scr.CurrentDesk))
-    return 0;
+  if (mask->my_flags.needs_current_page)
+  {
+    if (FScreenIsEnabled())
+    {
+      if (!FScreenIsRectangleOnScreen(NULL, FSCREEN_CURRENT, &(fw->frame_g)))
+	return 0;
+    }
+    else
+    {
+      if (!IsRectangleOnThisPage(&(fw->frame_g), Scr.CurrentDesk))
+	return 0;
+    }
+  }
+  else if (mask->my_flags.needs_current_global_page)
+  {
+    if (!IsRectangleOnThisPage(&(fw->frame_g), Scr.CurrentDesk))
+      return 0;
+  }
 
   /* Yes, I know this could be shorter, but it's hard to understand then */
   fMatchesName = matchWildcards(mask->name, fw->name);
@@ -333,12 +380,31 @@ Bool MatchesConditionMask(FvwmWindow *fw, WindowConditionMask *mask)
   if (mask->my_flags.needs_not_name && fMatches)
     return 0;
 
-  if ((mask->layer == -1) && (fw->layer != Scr.Focus->layer))
+  if ((mask->layer == -1) && (sf) && (fw->layer != sf->layer))
     return 0;
 
   if ((mask->layer >= 0) && (fw->layer != mask->layer))
     return 0;
 
+  if (mask->my_flags.use_do_accept_focus)
+  {
+    Bool f;
+
+    f = do_accept_input_focus(fw);
+    if (fw && HAS_NEVER_FOCUS(fw))
+    {
+      f = 0;
+    }
+    else if (fw && IS_LENIENT(fw))
+    {
+      f = 1;
+    }
+    if (!f != !mask->my_flags.do_accept_focus)
+    {
+      return 0;
+    }
+  }
+
   return 1;
 }
 
@@ -353,6 +419,7 @@ static FvwmWindow *Circulate(char *action, int Direction, char **restofline)
 {
   int pass = 0;
   FvwmWindow *fw, *found = NULL;
+  FvwmWindow *sf;
   WindowConditionMask mask;
   char *flags;
 
@@ -363,36 +430,45 @@ static FvwmWindow *Circulate(char *action, int Direction, char **restofline)
   { /* override for Current [] */
     mask.my_flags.use_circulate_hit = 1;
     mask.my_flags.use_circulate_hit_icon = 1;
+    mask.my_flags.use_circulate_hit_shaded = 1;
   }
   CreateConditionMask(flags, &mask);
   if (flags)
     free(flags);
-
-  if(Scr.Focus)
+  sf = get_focus_window();
+  if (sf)
   {
-    if(Direction == 1)
-      fw = Scr.Focus->prev;
-    else if(Direction == -1)
-      fw = Scr.Focus->next;
+    if (Direction > 0)
+      fw = sf->prev;
+    else if (Direction < 0)
+      fw = sf->next;
     else
-      fw = Scr.Focus;
+      fw = sf;
   }
   else
-    fw = NULL;
-
-  while((pass < 3)&&(found == NULL))
   {
-    while((fw)&&(found==NULL)&&(fw != &Scr.FvwmRoot))
+    if (Direction == 0)
+    {
+      return NULL;
+    }
+    fw = NULL;
+  }
+
+  for (pass = 0; pass < 3 && !found; pass++)
+  {
+    while (fw && !found && fw != &Scr.FvwmRoot)
     {
 #ifdef FVWM_DEBUG_MSGS
       fvwm_msg(DBG,"Circulate","Trying %s",fw->name);
 #endif /* FVWM_DEBUG_MSGS */
       /* Make CirculateUp and CirculateDown take args. by Y.NOMURA */
       if (MatchesConditionMask(fw, &mask))
+      {
 	found = fw;
+      }
       else
       {
-	if(Direction == 1)
+	if (Direction > 0)
 	  fw = fw->prev;
 	else
 	  fw = fw->next;
@@ -403,30 +479,29 @@ static FvwmWindow *Circulate(char *action, int Direction, char **restofline)
         return found;
       }
     }
-    if((fw == NULL)||(fw == &Scr.FvwmRoot))
+    if (!fw || fw == &Scr.FvwmRoot)
     {
-      if(Direction == 1)
+      if (Direction > 0)
       {
         /* Go to end of list */
-        fw = &Scr.FvwmRoot;
-        while(fw && fw->next)
-        {
-          fw = fw->next;
+	for (fw = &Scr.FvwmRoot; fw && fw->next; fw = fw->next)
+	{
+	  /* nop */
         }
       }
       else
       {
-        /* GO to top of list */
+        /* Go to top of list */
         fw = Scr.FvwmRoot.next;
       }
     }
-    pass++;
   }
   FreeConditionMask(&mask);
+
   return found;
 }
 
-void PrevFunc(F_CMD_ARGS)
+void CMD_Prev(F_CMD_ARGS)
 {
   FvwmWindow *found;
   char *restofline;
@@ -440,7 +515,7 @@ void PrevFunc(F_CMD_ARGS)
 
 }
 
-void NextFunc(F_CMD_ARGS)
+void CMD_Next(F_CMD_ARGS)
 {
   FvwmWindow *found;
   char *restofline;
@@ -454,7 +529,7 @@ void NextFunc(F_CMD_ARGS)
 
 }
 
-void NoneFunc(F_CMD_ARGS)
+void CMD_None(F_CMD_ARGS)
 {
   FvwmWindow *found;
   char *restofline;
@@ -467,7 +542,7 @@ void NoneFunc(F_CMD_ARGS)
   }
 }
 
-void CurrentFunc(F_CMD_ARGS)
+void CMD_Current(F_CMD_ARGS)
 {
   FvwmWindow *found;
   char *restofline;
@@ -480,7 +555,7 @@ void CurrentFunc(F_CMD_ARGS)
   }
 }
 
-void AllFunc(F_CMD_ARGS)
+void CMD_All(F_CMD_ARGS)
 {
   FvwmWindow *t, **g;
   char *restofline;
@@ -495,6 +570,7 @@ void AllFunc(F_CMD_ARGS)
     free(flags);
   mask.my_flags.use_circulate_hit = 1;
   mask.my_flags.use_circulate_hit_icon = 1;
+  mask.my_flags.use_circulate_hit_shaded = 1;
 
   num = 0;
   for (t = Scr.FvwmRoot.next; t; t = t->next)
@@ -537,7 +613,7 @@ static void GetDirectionReference(FvwmWindow *w, rectangle *r)
  * Execute a function to the closest window in the given
  * direction.
  **********************************************************************/
-void DirectionFunc(F_CMD_ARGS)
+void CMD_Direction(F_CMD_ARGS)
 {
   static char *directions[] =
   {
@@ -697,7 +773,7 @@ void DirectionFunc(F_CMD_ARGS)
 /* A very simple function, but handy if you want to call
  * complex functions from root context without selecting a window
  * for every single function in it. */
-void PickFunc(F_CMD_ARGS)
+void CMD_Pick(F_CMD_ARGS)
 {
   char *restofline;
   char *flags;
@@ -708,6 +784,9 @@ void PickFunc(F_CMD_ARGS)
 
   flags = CreateFlagString(action, &restofline);
   DefaultConditionMask(&mask);
+  mask.my_flags.use_circulate_hit = 1;
+  mask.my_flags.use_circulate_hit_icon = 1;
+  mask.my_flags.use_circulate_hit_shaded = 1;
   CreateConditionMask(flags, &mask);
   if (flags)
     free(flags);
@@ -716,12 +795,44 @@ void PickFunc(F_CMD_ARGS)
     old_execute_function(
       restofline, tmp_win, eventp, C_WINDOW, *Module, 0, NULL);
   }
+  FreeConditionMask(&mask);
 }
 
-void WindowIdFunc(F_CMD_ARGS)
+void CMD_ThisWindow(F_CMD_ARGS)
+{
+	char *restofline;
+	char *flags;
+	WindowConditionMask mask;
+
+	if (!tmp_win)
+	{
+		return;
+	}
+
+	flags = CreateFlagString(action, &restofline);
+	DefaultConditionMask(&mask);
+	mask.my_flags.use_circulate_hit = 1;
+	mask.my_flags.use_circulate_hit_icon = 1;
+	CreateConditionMask(flags, &mask);
+	if (flags)
+	{
+		free(flags);
+	}
+
+	if (MatchesConditionMask(tmp_win, &mask) && restofline)
+	{
+		old_execute_function(
+			restofline, tmp_win, eventp, C_WINDOW, *Module, 0,
+			NULL);
+	}
+
+	FreeConditionMask(&mask);
+	return;
+}
+
+void CMD_WindowId(F_CMD_ARGS)
 {
   FvwmWindow *t;
-  char *num;
   char *token;
   char *naction;
   unsigned long win;
@@ -731,31 +842,40 @@ void WindowIdFunc(F_CMD_ARGS)
   char *flags, *restofline;
 
   /* Get window ID */
-  action = GetNextToken(action, &num);
+  action = GetNextToken(action, &token);
 
-  if (num)
-  {
-    win = (unsigned long)strtol(num,NULL,0); /* SunOS doesn't have strtoul */
-    free(num);
-  }
-  else
-    win = 0;
-
-  token = PeekToken(action, &naction);
   if (token && StrEquals(token, "root"))
   {
     int screen = Scr.screen;
 
-    token = PeekToken(naction, NULL);
+    free(token);
+    token = PeekToken(action, &naction);
     if (!token || GetIntegerArguments(token, NULL, &screen, 1) != 1)
+    {
       screen = Scr.screen;
+    }
+    else
+    {
+      action = naction;
+    }
     use_screenroot = True;
     if (screen < 0 || screen >= Scr.NumberOfScreens)
       screen = 0;
-    w = XRootWindow(dpy, screen);
-    if (w == None)
+    win = XRootWindow(dpy, screen);
+    if (win == None)
       return;
   }
+  else if (token)
+  {
+    /* SunOS doesn't have strtoul */
+    win = (unsigned long)strtol(token, NULL, 0);
+    free(token);
+  }
+  else
+  {
+    win = 0;
+  }
+
   /* Look for condition - CreateFlagString returns NULL if no '(' or '[' */
   if (!use_screenroot)
   {
