@@ -430,6 +430,28 @@ static Bool must_move_transients(
   return False;
 }
 
+static Window __get_stacking_sibling(FvwmWindow *fw, Bool do_stack_below)
+{
+  Window w;
+
+  /* default to frame window */
+  w = fw->frame;
+  if (IS_ICONIFIED(fw) && do_stack_below == True)
+  {
+    /* override with icon windows when stacking below */
+    if (fw->icon_pixmap_w != None)
+    {
+      w = fw->icon_pixmap_w;
+    }
+    else if (fw->icon_w != None)
+    {
+      w = fw->icon_w;
+    }
+  }
+
+  return w;
+}
+
 static void restack_windows(
   FvwmWindow *r, FvwmWindow *s, int count, Bool do_broadcast_all, Bool do_lower)
 {
@@ -438,23 +460,23 @@ static void restack_windows(
   int i;
   XWindowChanges changes;
   Window *wins;
+  int do_stack_above;
+  int is_reversed;
 
   if (count <= 0)
   {
-    FvwmWindow *prev;
-
-    for (t = r, prev = NULL; prev != s; prev = t, t = t->stack_next)
+    for (count = 0, t = r->stack_next; t != s; t = t->stack_next)
     {
       count++;
       count += get_visible_icon_window_count(t);
     }
   }
   /* restack the windows between r and s */
-  wins = (Window*) safemalloc (count * sizeof (Window));
+  wins = (Window*) safemalloc ((count + 3) * sizeof (Window));
   i = 0;
   for (t = r->stack_next; t != s; t = t->stack_next)
   {
-    if (i >= count)
+    if (i > count)
     {
       fvwm_msg (ERR, "restack_windows", "more transients than expected");
       break;
@@ -469,19 +491,32 @@ static void restack_windows(
     }
   }
 
-  changes.sibling = (do_lower) ? r->frame : s->frame;
-  if (changes.sibling != None)
+  changes.sibling = __get_stacking_sibling(r, True);
+  if (changes.sibling == None)
   {
-    changes.stack_mode = (do_lower) ? Below : Above;
-    flags = CWSibling|CWStackMode;
+    changes.sibling = __get_stacking_sibling(s, False);
+    is_reversed = 1;
   }
   else
   {
-    changes.stack_mode = (do_lower) ? Above : Below;
+    is_reversed = 0;
+  }
+  if (changes.sibling == None)
+  {
+    do_stack_above = !do_lower;
     flags = CWStackMode;
   }
+  else
+  {
+    do_stack_above = 0;
+    flags = CWStackMode | CWSibling;
+  }
+  changes.stack_mode = (do_stack_above ^ is_reversed) ? Above : Below;
   XConfigureWindow (dpy, r->stack_next->frame, flags, &changes);
-  XRestackWindows (dpy, wins, count);
+  if (count > 1)
+  {
+    XRestackWindows (dpy, wins, count);
+  }
   free(wins);
   if (do_broadcast_all)
   {
@@ -561,9 +596,7 @@ static Bool __restack_window(
   /* detach t, so it doesn't make trouble in the loops */
   remove_window_from_stack_ring(t);
 
-  count = 1;
-  count += get_visible_icon_window_count(t);
-
+  count = 0;
   if (do_restack_transients)
   {
     /* collect the transients in a temp list */
@@ -574,8 +607,8 @@ static Bool __restack_window(
     {
       do_restack_transients = False;
     }
-    count++;
   }
+  count += 1 + get_visible_icon_window_count(t);
 
   test_layer = t->layer;
   if (do_lower)
@@ -661,12 +694,12 @@ static void __RaiseOrLowerWindow(
       for (t2 = Scr.FvwmRoot.stack_next; t2 != &Scr.FvwmRoot;
 	   t2 = t2->stack_next)
       {
-	if (t2 == t)
-	{
-	  break;
-	}
         if (t2->w == t->transientfor)
         {
+	  if (t2 == t)
+	  {
+	    break;
+	  }
 	  if (IS_ICONIFIED(t2) || t->layer != t2->layer)
 	  {
 	    break;
@@ -1026,9 +1059,12 @@ static void ResyncXStackingOrder(void)
 /* send RESTACK packets for all windows between s1 and s2 */
 static void BroadcastRestack(FvwmWindow *s1, FvwmWindow *s2)
 {
-  FvwmWindow *t, *t2;
-  int num, i;
+  FvwmWindow *fw;
+  int num;
+  int i;
+  int n;
   unsigned long *body, *bp, length;
+  unsigned long max_wins_per_packet;
   extern Time lastTimestamp;
 
   if (s2 == &Scr.FvwmRoot)
@@ -1039,46 +1075,51 @@ static void BroadcastRestack(FvwmWindow *s1, FvwmWindow *s2)
   }
   if (s1 == &Scr.FvwmRoot)
   {
-    t = s1->stack_next;
-    if (t == &Scr.FvwmRoot)
+    s1 = s1->stack_next;
+    if (s1 == &Scr.FvwmRoot)
       return;
-    /* t has been moved to the top of stack */
-    BroadcastPacket (M_RAISE_WINDOW, 3, t->w, t->frame, (unsigned long)t);
-    if (t->stack_next == s2)
+    /* s1 has been moved to the top of stack */
+    BroadcastPacket (M_RAISE_WINDOW, 3, s1->w, s1->frame, (unsigned long)s1);
+    if (s1->stack_next == s2)
     {
       /* avoid sending empty RESTACK packet */
       return;
     }
   }
-  else
-  {
-    t = s1;
-  }
   if (s1 == s2)
   {
     /* A useful M_RESTACK packet must contain at least two windows. */
     return;
   }
-  for (t2 = t, num = 1 ; t2 != s2 && t != &Scr.FvwmRoot;
-       t2 = t2->stack_next, num++)
-    ;
-  length = FvwmPacketHeaderSize + 3*num;
-  body = (unsigned long *) safemalloc (length*sizeof(unsigned long));
-
-  bp = body;
-  *(bp++) = START_FLAG;
-  *(bp++) = M_RESTACK;
-  *(bp++) = length;
-  *(bp++) = lastTimestamp;
-  for (t2 = t; num != 0; num--, t2 = t2->stack_next)
+  for (fw = s1, num = 1; fw != s2 && fw != &Scr.FvwmRoot;
+       fw = fw->stack_next, num++)
   {
-    *(bp++) = t2->w;
-    *(bp++) = t2->frame;
-    *(bp++) = (unsigned long)t2;
+	  /* nothing */
+  }
+  max_wins_per_packet = (FvwmPacketMaxSize - FvwmPacketHeaderSize) / 3;
+  /* split packet if it is too long */
+  for ( ; num > 1; s1 = fw, num -= n)
+  {
+    n = min(num, max_wins_per_packet) - 1;
+    length = FvwmPacketHeaderSize + 3 * (n + 1);
+    body = (unsigned long *)safemalloc(length * sizeof(unsigned long));
+    bp = body;
+    *(bp++) = START_FLAG;
+    *(bp++) = M_RESTACK;
+    *(bp++) = length;
+    *(bp++) = lastTimestamp;
+    for (fw = s1, i = 0; i <= n; i++, fw = fw->stack_next)
+    {
+      *(bp++) = fw->w;
+      *(bp++) = fw->frame;
+      *(bp++) = (unsigned long)fw;
+    }
+    for (i = 0; i < npipes; i++)
+    {
+      PositiveWrite(i, body, length*sizeof(unsigned long));
+    }
+    free(body);
   }
-  for (i = 0; i < npipes; i++)
-    PositiveWrite(i, body, length*sizeof(unsigned long));
-  free(body);
 #ifdef DEBUG_STACK_RING
   verify_stack_ring_consistency();
 #endif
@@ -1203,9 +1244,9 @@ void mark_transient_subtree(
       }
       if (!IS_TRANSIENT(s) && !use_group_hint)
 	continue;
-      r = (FvwmWindow *)s->scratch.p;
-      if (do_ignore_icons && r && IS_ICONIFIED(r))
+      if (do_ignore_icons && IS_ICONIFIED(s))
 	continue;
+      r = (FvwmWindow *)s->scratch.p;
       if (IS_TRANSIENT(s))
       {
 	if (r && IS_IN_TRANSIENT_SUBTREE(r) &&
