@@ -75,6 +75,22 @@ static struct charstring key_modifiers[]=
   {0,0}
 };
 
+/* Table to translate a modifier map index to a modifier that we define that
+ * generates that index.  This mapping can be chosen by each client, but the
+ * settings below try to emulate the usual terminal behaviour. */
+static unsigned int modifier_mapindex_to_mask[8] =
+{
+	ShiftMask,
+	Mod3Mask, /* Alt Gr */
+	Mod3Mask | ShiftMask,
+	/* Just guessing below here - LockMask is not used anywhere*/
+	ControlMask,
+	Mod1Mask, /* Alt/Meta on XFree86 */
+	Mod2Mask, /* Num lock on XFree86 */
+	Mod4Mask,
+	Mod5Mask, /* Scroll lock on XFree86 */
+};
+
 static int key_min;
 static int key_max;
 
@@ -262,7 +278,7 @@ Bool MatchBinding(Display *dpy, Binding *b,
   static int kc_maxlen = 0;
   static KeySym last_keysym = NoSymbol;
 
-  modifier &= used_modifiers;
+  modifier &= (used_modifiers & ALL_MODIFIERS);
   if (((b->Modifier & used_modifiers) != modifier &&
        b->Modifier != AnyModifier) ||
       (b->Context & Context) == 0 ||
@@ -349,6 +365,7 @@ int AddBinding(
   int max;
   int maxmods;
   int m;
+  int mask;
   int count = 0;
   KeySym tkeysym;
   Binding *temp;
@@ -388,6 +405,7 @@ int AddBinding(
       {
 	unsigned int add_modifiers = 0;
 	unsigned int bind_mask = 1;
+	unsigned int check_bound_mask = 0;
 
 	switch (m)
 	{
@@ -400,22 +418,25 @@ int AddBinding(
 	  {
 	    add_modifiers = ShiftMask;
 	    bind_mask = 2;
-	  }
-	  break;
-	case 2:
-	  /* key generates the key sym with caps-lock depressed */
-	  if (modifiers != AnyModifier && !(modifiers & LockMask))
-	  {
-	    add_modifiers = LockMask;
-	    bind_mask = 2;
+	    /* but don't bind it again if already bound without modifiers */
+	    check_bound_mask = 1;
 	  }
 	  break;
 	default:
-	  /* key generates the key sym with unknown modifiers depressed -
-	  * can't map that to specific modifiers - trat as no modifiers */
+	  /* key generates the key sym with undefined modifiers depressed -
+	   * let's make an educated guess at what modifiers the user expected
+	   * based on the XFree86 default configuration. */
+	  mask = modifier_mapindex_to_mask[m - 1];
+	  if (modifiers != AnyModifier && !(modifiers & mask) != mask)
+	  {
+	    add_modifiers = mask;
+	    bind_mask = (1 << m);
+	    /* but don't bind it again if already bound without modifiers */
+	    check_bound_mask = 1;
+	  }
 	  break;
 	}
-	if (bind_mask & bound_mask)
+	if ((bind_mask & bound_mask) || (check_bound_mask & bound_mask))
 	{
 	  /* already bound, break out */
 	  break;
@@ -545,7 +566,7 @@ void *CheckBinding(Binding *blist,
   Binding *b;
   unsigned int used_modifiers = ~dead_modifiers;
 
-  modifier &= used_modifiers;
+  modifier &= (used_modifiers & ALL_MODIFIERS);
 
   for (b = blist; b != NULL; b = b->NextBinding)
   {
@@ -574,6 +595,7 @@ Bool MatchBindingExactly(
   int button, KeyCode keycode, unsigned int modifier, int Context,
   BindingType type)
 {
+  modifier &= ALL_MODIFIERS;
   if (b->type == type)
   {
     if (
@@ -620,6 +642,7 @@ void GrabWindowKey(Display *dpy, Window w, Binding *binding,
 {
   /* remove unnecessary bits from dead_modifiers */
   dead_modifiers &= ~(binding->Modifier & dead_modifiers);
+  dead_modifiers &= ALL_MODIFIERS;
 
   if((binding->Context & contexts) && (binding->type == KEY_BINDING))
   {
@@ -678,6 +701,7 @@ void GrabWindowButton(Display *dpy, Window w, Binding *binding,
     return;
 
   dead_modifiers &= ~(binding->Modifier & dead_modifiers); /* dje */
+  dead_modifiers &= ALL_MODIFIERS;
 
   if ((binding->Context & contexts) &&
       ((binding->type == MOUSE_BINDING)
@@ -692,11 +716,15 @@ void GrabWindowButton(Display *dpy, Window w, Binding *binding,
     for (button = bmin; button <= bmax; button++)
     {
       if (fGrab)
+      {
 	XGrabButton(dpy, button, binding->Modifier, w,
 		    True, ButtonPressMask | ButtonReleaseMask,
 		    GrabModeSync, GrabModeAsync, None, cursor);
+      }
       else
+      {
 	XUngrabButton(dpy, button, binding->Modifier, w);
+      }
       if(binding->Modifier != AnyModifier && dead_modifiers != 0)
       {
 	register unsigned int mods;
@@ -770,6 +798,27 @@ void GrabAllWindowKeysAndButtons(Display *dpy, Window w, Binding *blist,
   return;
 }
 
+void GrabWindowKeyOrButton(
+  Display *dpy, Window w, Binding *binding, unsigned int contexts,
+  unsigned int dead_modifiers, Cursor cursor, Bool fGrab)
+{
+  switch (binding->type)
+  {
+  case MOUSE_BINDING:
+  STROKE_CODE(case STROKE_BINDING:)
+    GrabWindowButton(
+      dpy, w, binding, contexts, dead_modifiers, cursor, fGrab);
+    break;
+  case KEY_BINDING:
+    GrabWindowKey(dpy, w, binding, contexts, dead_modifiers, fGrab);
+    break;
+  default:
+    break;
+  }
+
+  return;
+}
+
 /***********************************************************************
  *
  *  Procedure:
